/**
 * TERMS AND CONDITIONS:
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.backup.zipsnap.ng.job;

import java.io.File;
import java.io.IOException;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.nio.charset.Charset;
import java.util.Date;
import java.util.HashMap;

import com.backup.zipsnap.ng.Main;
import com.backup.zipsnap.ng.gui.JobWrapper.msg;
import com.backup.zipsnap.ng.gui.JobWrapper;
import com.backup.zipsnap.ng.gui.Tab;
import com.backup.zipsnap.ng.job.util.FileIO;
import com.backup.zipsnap.ng.job.util.Logger;
import com.backup.zipsnap.ng.job.util.Timer;
import com.backup.zipsnap.ng.job.util.UserIO;
import com.backup.zipsnap.ng.job.util.UserIO.out;
import com.javamex.classmexer.MemoryUtil;
import com.javamex.classmexer.MemoryUtil.VisibilityFilter;


public class Job {

	/* configuration file and class */
	public static ThreadLocal<File> configFile = new ThreadLocal<File>();
	public static ThreadLocal<Config> config = new ThreadLocal<Config>();

	/* catalog file and class */
	public static ThreadLocal<Catalog> catalog = new ThreadLocal<Catalog>();

	/* current timestamp */
	public static ThreadLocal<String> timestamp = new ThreadLocal<String>(); 
	public static ThreadLocal<Date> currentTime = new ThreadLocal<Date>();

	/* global variables */
	public static ThreadLocal<Timer> timer = new ThreadLocal<Timer>();
	public static ThreadLocal<Logger> logger = new ThreadLocal<Logger>();
	public static ThreadLocal<String> command = new ThreadLocal<String>();
	public static ThreadLocal<Tab> thisTab = new ThreadLocal<Tab>();
	public static boolean isInGUI = false;

	/*
	 * Constructor
	 */
	public Job(boolean isInGUI) {
		Job.isInGUI = isInGUI;
	}

	/*
	 * run a job
	 */
	public void run(final Tab tab,final String[] args) {
		currentTime.set(new Date());
		timer.set(new Timer());
		logger.set(new Logger());	
		thisTab.set(tab);
		UserIO.warnings.set(new HashMap<String,out>());
		Job.gui(JobWrapper.msg.STATUS,"started");
		/* load an empty config and catalog */
		config.set(new Config());
		catalog.set(new Catalog());
		timestamp.set(config.get().timestampFormatter.format(currentTime.get()));

		/* check if the Zip4j class is loaded */
		try {
			Class.forName("net.lingala.zip4j.core.ZipFile");
		} catch (Exception e) {
			UserIO.reportError("Unable to load the required Zip4j library");
		}

		/* if running in the GUI, pass STDIN to it */
		if (Job.isInGUI) {
			PipedOutputStream inwriter = new PipedOutputStream();
			PipedInputStream input = null;
			try {
				input = new PipedInputStream(inwriter);
			} catch (IOException e) {
				e.printStackTrace();
			}
			System.setIn(input);
			Job.gui(JobWrapper.msg.STDIN,inwriter);
		} 

		/* print program title */
		timer.get().start("JOB");
		UserIO.print(UserIO.out.INFO,Main.programTitle+Main.lineSeparator+"=================================");

		try {

			/**********************************
			 * (1) READ COMMAND LINE COMMANDS *
			 *********************************/
			/* no command line arguments are provided, exit */
			if (args.length == 0) {
				printUsage();
				System.exit(0);
			}  
			if (args.length < 2) UserIO.reportError("Wrong number of arguments: at least a configuration file and a command must be provided");
			UserIO.print(UserIO.out.VERBOSE,"Configuration file: "+args[0]+", Command: "+args[1]);

			/* configuration file is the first argument */
			configFile.set(new File(args[0]));

			/* command is the second argument */
			String cmd = args[1];
			if ("backup".equals(cmd)) command.set("backup");
			else if ("restore".equals(cmd)) command.set("restore");
			else if ("info".equals(cmd)) command.set("info");
			else if ("purge".equals(cmd)) command.set("purge");
			else if ("config".equals(cmd)) command.set("config");
			else UserIO.reportError("Invalid command (" + command.get() +")");
			Job.gui(JobWrapper.msg.CONFIG,config.get());

			/* process other arguments, if any */
			if ((command.get().equals("restore")||command.get().equals("info")) && args.length == 3) {
				/* snapshot number to be restored or for info */
				try {
					config.get().snapshot = Integer.parseInt(args[2]);
				} catch (Exception e) {
					UserIO.reportError("Invalid snapshot provided \"" + args[2] + "\": snapshot number must be a positive integer");
				}
			} 
			if (command.get().equals("backup") && args.length == 3) {
				/* process the path of the shadow volume */
				config.get().shadowDevice = args[2];
			}
			if (command.get().equals("config")) {
				config.get().generate(configFile.get(),true);
			}

			if (!configFile.get().exists()) UserIO.reportError("The configuration file provided (" + configFile.get().getPath() + ") does not exist");
			/*********************************
			 * (2) PARSE CONFIGURATION FILE *
			 ********************************/
			config.get().parse(configFile.get());
			memoryUsage("Config",config.get());
			Job.gui(JobWrapper.msg.JOB_CONFIG,config.get());	
			/* print operating system and java version details */
			UserIO.print(UserIO.out.DEBUG,"Operating System: "+System.getProperty("os.name")+" ("+System.getProperty("os.version")+" on "+System.getProperty("os.arch")+")");
			UserIO.print(UserIO.out.DEBUG,"Java version: "+System.getProperty("java.version")+" ("+System.getProperty("java.vendor")+")");
			UserIO.print(UserIO.out.DEBUG,"Charset Encoding: "+Charset.defaultCharset().displayName());

			/***************************************
			 * (3) CHECK THE ARCHIVE DIRECTORY      *
			 ***************************************/

			/* check if the archive directory is valid */
			FileIO.checkDirectory(config.get().archive);
			UserIO.print(UserIO.out.INFO,"Archive directory: " + config.get().archive.nativeName);

			/***************************************
			 * (4) LOAD THE CATALOG                *
			 ***************************************/
			UserIO.print(UserIO.out.INFO,"Loading the catalog...",false);
			timer.get().start("LOADING CATALOG");
			catalog.get().load();
			UserIO.print(UserIO.out.INFO,"done");
			UserIO.print(UserIO.out.DEBUG,"Elapsed: "+timer.get().stop("LOADING CATALOG"));
			memoryUsage("Catalog",catalog.get());

			/***************************************
			 * (5) EXECUTE REQUESTED COMMAND       *
			 ***************************************/

			if (command.get().equals("backup")) {
				/* new snapshot is +1 of the last snapshot */
				config.get().snapshot = catalog.get().snapshotsIndex.size()+1;
				logger.get().openLog(new File(config.get().archive.file, FileIO.getCustomFilename("log-backup") + ".txt"));
				Backup.doBackup();
			}
			else if (command.get().equals("restore")) {
				/* if snapshot is not set, restore the last one */
				if (config.get().snapshot == -1) config.get().snapshot = catalog.get().snapshotsIndex.size();
				logger.get().openLog(new File(config.get().archive.file, FileIO.getCustomFilename("log-restore") + ".txt"));
				Restore.doRestore();
			}
			else if (command.get().equals("purge")) {
				config.get().snapshot = catalog.get().snapshotsIndex.size();
				logger.get().openLog(new File(config.get().archive.file, FileIO.getCustomFilename("log-purge") + ".txt"));
				Purge.doPurge(true);
			}
			else if (command.get().equals("info")) {
				if (config.get().snapshot > 0) {
					/* the user requested to show the info a specific snapshot */
					logger.get().openLog(new File(config.get().archive.file, FileIO.getCustomFilename("log-info-snapshot") + ".txt"));
					Info.snapshotInfo();
				} else {
					/* provide a summary of all the snapshots */
					config.get().snapshot = catalog.get().snapshotsIndex.size();
					logger.get().openLog(new File(config.get().archive.file, FileIO.getCustomFilename("log-info-archive") + ".txt"));
					Info.archiveSummary(true);
				}
			}

			/***************************************
			 * (6) PRINT REPORTED WARNINGS          *
			 ***************************************/

			if (Main.settings.printWarningsSummary && UserIO.warnings.get().size() > 0) {
				UserIO.print(UserIO.out.INFO,Main.lineSeparator+"[WARNINGS]"+Main.lineSeparator+"The following warnings have been reported:");
				for ( String message : UserIO.warnings.get().keySet() ) UserIO.print(UserIO.warnings.get().get(message),"- "+message);
			}
			/* catch all the exceptions */
		} catch (Exception e) {
			UserIO.reportError("An unexpected error has occurred:" +Main.lineSeparator+ UserIO.getExceptionMessage(e));
		}

		/***************************************
		 * (7) EXIT                            *
		 ***************************************/
		Job.exit(0);
	}

	/*
	 * Print out usage syntax, notes, and comments.
	 */
	public static void printUsage() {
		/* RULER   00000000011111111112222222222333333333344444444445555555555666666666677777777778 */
		/* RULER   12345678901234567890123456789012345678901234567890123456789012345678901234567890 */
		UserIO.print(UserIO.out.INFO,
				"ZipSnapNG is an enhanced incremental backup tool"+
						Main.lineSeparator+""+
						Main.lineSeparator+"GUI USAGE:  ZipSnapNG gui"+
						Main.lineSeparator+"CLI USAGE:  ZipSnapNG <config_file> <command> [option]"+
						Main.lineSeparator+""+
						Main.lineSeparator+"<config_file>: the location of the configuration file"+
						Main.lineSeparator+""+
						Main.lineSeparator+"<command>:"+
						Main.lineSeparator+" backup      Add a snapshot of the directories defined in the configuration"+
						Main.lineSeparator+"             file into the archive. Only new or modified files are added to" +
						Main.lineSeparator+"             the catalog and only the portion that has changed is saved." +
						Main.lineSeparator+" restore     Restore the last snapshot from the archive to the original"+
						Main.lineSeparator+"             location or into a different directory, as specified in the" +
						Main.lineSeparator+"             configuration file. For updated items, only the portion of the" +
						Main.lineSeparator+"             files that needs to be restored is extracted from the archive." +
						Main.lineSeparator+" info        Show information regarding the archive pointed by the"+
						Main.lineSeparator+"             configuration file, including number of files and size." +
						Main.lineSeparator+" purge       Apply the retention policies set in the configuration file to the"+
						Main.lineSeparator+"             catalog." +
						Main.lineSeparator+" config      Generate a sample configuration file at the location specified by"+
						Main.lineSeparator+"             <config_file>" +
						Main.lineSeparator+"" +
						Main.lineSeparator+"[option]:" +
						Main.lineSeparator+" For \"backup\" is the path of the shadow copy device" +
						Main.lineSeparator+" For \"restore\" is the snapshot number to restore" +
						Main.lineSeparator+" For \"info\" is the snapshot you want to list all the items" +
						Main.lineSeparator+""
				);
	}

	/*
	 * calculate memory usage of a given object 
	 */
	public static void memoryUsage(String name, Object o) {
		if (Main.settings.memoryStats) 
			UserIO.print(UserIO.out.DEBUG,"(MEMORY-STATS) "+name+": " + FileIO.getSize(MemoryUtil.deepMemoryUsageOf(o,VisibilityFilter.ALL)));
	}
	
	/*
	 * exit the program
	 */
	@SuppressWarnings("deprecation")
	public static void exit(int exitCode){
		String completedIn = timer.get().stop("JOB");
		UserIO.print(UserIO.out.INFO,"Completed in "+completedIn);
		Job.gui(JobWrapper.msg.FINISHED, completedIn);
		memoryUsage("Timer",timer.get());
		memoryUsage("Logger",logger.get());
		logger.get().closeLog();
		Config.initialized.set(new Boolean(false));

		if (isInGUI) {
			/* notify the gui the job is finished */
			Job.gui(JobWrapper.msg.STATUS,"finished");
			if (exitCode == 0) return;
			else Thread.currentThread().stop();
		} else System.exit(exitCode);
	}
	
	/*
	 * send a message back to the gui
	 */
	public static void gui(msg what, Object obj) {
		if (thisTab.get() != null) thisTab.get().jobEngine.callback(what,obj);
	}
}


